import {
  Box,
  Divider,
  MenuItem,
  Stack,
  TextField,
  Typography,
} from '@mui/material'
import { Controller, SubmitHandler, useForm } from 'react-hook-form'
import { DatePicker } from '@mui/lab'
import React, {
  forwardRef,
  Ref,
  useContext,
  useEffect,
  useImperativeHandle,
  useState,
} from 'react'
import { addDays } from 'date-fns/fp'
import { useTranslation } from 'react-i18next'
import { RegisterFormData } from './formData'
import { PersonFormMethods } from './personFormMethods'
import useOus, { enSort, nbSort } from '../../../hooks/useOus'
import useRoleTypes, { RoleTypeData } from '../../../hooks/useRoleTypes'
import { isValidEmail } from '../../../utils'
import { FeatureContext } from '../../../contexts'

interface StepPersonFormProperties {
  nextHandler(formState: RegisterFormData): void

  formData: RegisterFormData
}

const StepPersonForm = forwardRef(
  (props: StepPersonFormProperties, ref: Ref<PersonFormMethods>) => {
    const { i18n, t } = useTranslation(['common'])
    const { nextHandler, formData } = props
    const ous = useOus()
    const [ouChoice, setOuChoice] = useState(
      formData.ou_id ? formData.ou_id : ''
    )
    const [selectedRoleType, setSelectedRoleType] = useState(
      formData.role_type ? formData.role_type : ''
    )
    const today = new Date()
    const [todayPlusMaxDays, setTodayPlusMaxDays] = useState(today)
    const roleTypes = useRoleTypes()
    const { displayContactAtUnit, displayComment } = useContext(FeatureContext)

    const roleTypeSort = () => (a: RoleTypeData, b: RoleTypeData) => {
      if (i18n.language === 'en') {
        return a.name_nb.localeCompare(b.name_nb)
      }
      return a.name_en.localeCompare(b.name_en)
    }

    const submit: SubmitHandler<RegisterFormData> = (data) => {
      nextHandler(data)
    }

    // Set up the form with the data given as input to the component
    const {
      register,
      control,
      handleSubmit,
      formState: { errors },
      setValue,
      getValues,
      trigger,
    } = useForm<RegisterFormData>({ defaultValues: formData })
    const onSubmit = handleSubmit(submit)

    const handleOuChange = (event: any) => {
      if (event.target.value) {
        setOuChoice(event.target.value)
        setValue('ou_id', parseInt(event.target.value, 10))
      } else {
        setValue('ou_id', undefined)
      }
    }

    register('role_type', {
      required: t<string>('validation.roleTypeRequired'),
    })

    const getMaxDateForRoleType = (roleType: RoleTypeData | undefined) => {
      if (roleType === undefined) {
        return today
      }
      return addDays(roleType.max_days)(today)
    }

    function updateMaxDaysForRoleType(roleTypeId: any) {
      const selectedRoleTypeInfo = roleTypes.find((rt) => rt.id === roleTypeId)
      const maxDate = getMaxDateForRoleType(selectedRoleTypeInfo)
      setTodayPlusMaxDays(maxDate)
      // Trigger revalidation of role end since the max end date is
      // different for different role types
      trigger('role_end')
    }

    // When the component is first rendered and there is some role_type set in
    // the input formData, the roleTypes-array is most
    // probably empty since the data has not come back from the server yet.
    // Need to have this effect here to make sure the max end date gets
    // updated when the role types are available
    useEffect(() => {
      if (formData.role_type) {
        updateMaxDaysForRoleType(formData.role_type)
      }
    }, [roleTypes])

    const handleRoleTypeChange = (event: any) => {
      if (event.target.value) {
        setValue('role_type', event.target.value)
        setSelectedRoleType(event.target.value)
        updateMaxDaysForRoleType(event.target.value)
      } else {
        setValue('role_type', undefined)
      }
    }

    function doSubmit() {
      return onSubmit()
    }

    useImperativeHandle(ref, () => ({ doSubmit }))

    const validateStartDateBeforeEndDate = (startDate: Date | undefined) => {
      if (!startDate) {
        return t('validation.startDateMustBeSet')
      }

      const roleEnd = getValues('role_end')
      if (roleEnd && startDate > roleEnd) {
        // The role end date is set, but is is before the start date
        return t('validation.startDateMustBeBeforeEndDate')
      }
      return true
    }

    return (
      <>
        <Typography
          variant="h5"
          sx={{
            paddingTop: '1rem',
            paddingBottom: '1rem',
          }}
        >
          {t('registerWizardText.registerContactInfo')}
        </Typography>
        <Typography sx={{ paddingBottom: '0.5rem' }}>
          {t('registerWizardText.registerPage')}
        </Typography>
        <Divider sx={{ border: '1px solid' }} />
        <form onSubmit={onSubmit}>
          <Box
            sx={{
              paddingTop: '1rem',
              paddingBottom: '4rem',
              maxWidth: '30rem',
            }}
          >
            <Stack spacing={2}>
              <TextField
                id="firstName"
                label={t('input.firstName')}
                error={!!errors.first_name}
                helperText={errors.first_name && errors.first_name.message}
                {...register(`first_name`, {
                  required: t<string>('validation.firstNameRequired'),
                })}
              />
              <TextField
                id="lastName"
                label={t('input.lastName')}
                error={!!errors.last_name}
                helperText={errors.last_name && errors.last_name.message}
                {...register(`last_name`, {
                  required: t<string>('validation.lastNameRequired'),
                })}
                inputProps={{ 'data-testid': 'lastName-input-field' }}
              />
              <TextField
                id="email"
                label={t('input.email')}
                error={!!errors.email}
                helperText={errors.email && errors.email.message}
                {...register(`email`, {
                  validate: isValidEmail,
                })}
              />
            </Stack>
          </Box>
          <Typography
            variant="h5"
            sx={{
              paddingTop: '1rem',
              paddingBottom: '1rem',
            }}
          >
            {t('registerWizardText.registerPeriod')}
          </Typography>
          <Typography sx={{ paddingBottom: '0.5rem' }}>
            {t('registerWizardText.registerPeriodText')}
          </Typography>
          <Divider sx={{ border: '1px solid' }} />

          <Box sx={{ paddingTop: '1rem', maxWidth: '30rem' }}>
            <Stack spacing={2}>
              <TextField
                id="roletype-select"
                select
                value={selectedRoleType}
                error={!!errors.role_type}
                label={t('input.roleType')}
                onChange={handleRoleTypeChange}
              >
                {roleTypes.sort(roleTypeSort()).map((roleType) => (
                  <MenuItem key={roleType.id.toString()} value={roleType.id}>
                    {i18n.language === 'en'
                      ? roleType.name_en
                      : roleType.name_nb}
                  </MenuItem>
                ))}
              </TextField>

              <TextField
                id="ou-select-label"
                select
                value={ouChoice?.toString()}
                label={t('common:ou')}
                onChange={handleOuChange}
              >
                {ous
                  .sort(i18n.language === 'en' ? enSort : nbSort)
                  .map((ou) => (
                    <MenuItem key={ou.id.toString()} value={ou.id}>
                      {i18n.language === 'en' ? ou.en : ou.nb} ({ou.id})
                    </MenuItem>
                  ))}
              </TextField>

              {/* There are no particular constraints on the date pickers. It should be allowed to add a role with a start date that is in the past for instance */}
              <Controller
                name="role_start"
                control={control}
                rules={{ validate: validateStartDateBeforeEndDate }}
                defaultValue={today}
                render={({ field }) => (
                  <DatePicker
                    mask="____-__-__"
                    label={t('input.roleStartDate')}
                    maxDate={todayPlusMaxDays}
                    value={field.value}
                    inputFormat="yyyy-MM-dd"
                    onChange={(value) => {
                      field.onChange(value)
                    }}
                    renderInput={(params) => <TextField {...params} />}
                  />
                )}
              />
              {!!errors.role_start && (
                <Box sx={{ typography: 'caption', color: 'red' }}>
                  {errors.role_start.message}
                </Box>
              )}
              <Controller
                name="role_end"
                control={control}
                defaultValue={today}
                render={({ field }) => (
                  <DatePicker
                    mask="____-__-__"
                    label={t('input.roleEndDate')}
                    maxDate={todayPlusMaxDays}
                    value={field.value}
                    inputFormat="yyyy-MM-dd"
                    onChange={(value) => {
                      field.onChange(value)
                    }}
                    renderInput={(params) => <TextField {...params} />}
                  />
                )}
              />

              {displayContactAtUnit && (
                <TextField
                  id="contact_person"
                  label={`${t('input.contactPersonUnit')} ${t(
                    'input.optional'
                  )}`}
                  {...register(`contact_person_unit`)}
                />
              )}

              {displayComment && (
                <TextField
                  id="comment"
                  label={`${t('input.comment')} ${t('input.optional')}`}
                  multiline
                  rows={5}
                  {...register(`comment`)}
                />
              )}
            </Stack>
          </Box>
        </form>
      </>
    )
  }
)

export default StepPersonForm
