"""
Fetch all OUs from OrgReg and add the complete tree to Greg.

Ignores OrganizationalUnits without identifiers with source and name matching global
variables ORGREG_SOURCE and ORGREG_NAME

Assumes that the header used for authentication is of the type
'X-Gravitee-Api-Key': 'token'.

If the path to the endpoint of the OUs is oregreg/v3/ou/ you want to give
orgreg/v3/ as the url argument (note the trailing slash).
"""
import datetime
import logging
from typing import Union, Mapping, Dict, Optional

import orgreg_client
from django.conf import settings
from django.core.management.base import BaseCommand
from orgreg_client import OrgUnit
from orgreg_client.models import ExternalKey

from greg.models import OrganizationalUnit, OuIdentifier

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = __doc__
    processed: Dict[int, OrganizationalUnit] = {}

    def _upsert_extra_identities(
        self,
        ou: OrgUnit,
    ):
        """Upsert any configured extra IDs from orgreg."""
        for extra_id in settings.ORGREG_EXTRA_IDS:
            identity_in_orgreg = self._get_external_key_from_ou(extra_id, ou)
            if identity_in_orgreg is not None:
                self._upsert_identifier(
                    extra_id["source"],
                    extra_id["type"],
                    identity_in_orgreg.value,
                    ou.ou_id,
                )

        # Acronyms can also be used as identifiers
        for acronym in settings.ORGREG_ACRONYMS:
            self.handle_acronym_identifier(acronym, ou)

    def handle_acronym_identifier(self, acronym, ou):
        if acronym == "nob" and ou.acronym.nob is not None:
            self._upsert_identifier(
                settings.ORGREG_SOURCE,
                "acronym_" + acronym,
                ou.acronym.nob,
                ou.ou_id,
            )
        if acronym == "eng" and ou.acronym.eng is not None:
            self._upsert_identifier(
                settings.ORGREG_SOURCE,
                "acronym_" + acronym,
                ou.acronym.eng,
                ou.ou_id,
            )
        if acronym == "nno" and ou.acronym.nno is not None:
            self._upsert_identifier(
                settings.ORGREG_SOURCE,
                "acronym_" + acronym,
                ou.acronym.nno,
                ou.ou_id,
            )

    @staticmethod
    def _get_external_key_from_ou(
        extra_id: Dict[str, str], ou: OrgUnit
    ) -> Optional[ExternalKey]:
        matching_ids = [
            x
            for x in ou.external_keys
            if x.source_system == extra_id["source"] and x.type == extra_id["type"]
        ]

        if not matching_ids:
            logger.warning(
                "No %s id from %s found in OrgReg for ou %s",
                extra_id["type"],
                extra_id["source"],
                ou.ou_id,
            )
            return None

        if len(matching_ids) > 1:
            # External_ids
            logger.warning(
                "Found multiple ids matching type: %s source: %s in OrgReg. Using the first one: %s",
                extra_id["type"],
                extra_id["source"],
                matching_ids[0].value,
            )

        return matching_ids[0]

    def _upsert_identifier(
        self, source: str, identity_type: str, identity_in_orgreg_value: str, ou_id: int
    ):
        # Check if the id exists
        identify_in_db = (
            self.processed[ou_id]
            .identifiers.filter(
                source=source,
                name=identity_type,
            )
            .first()
        )

        if identify_in_db:
            if identify_in_db.value != identity_in_orgreg_value:
                logger.info(
                    "Updating id: source: %s, type: %s, old_id: %s, new_id %s",
                    source,
                    identity_type,
                    identify_in_db.value,
                    identity_in_orgreg_value,
                )
                identify_in_db.value = identity_in_orgreg_value
                identify_in_db.save()
        else:
            OuIdentifier.objects.create(
                name=identity_type,
                source=source,
                value=identity_in_orgreg_value,
                orgunit=self.processed[ou_id],
            )
            logger.info(
                "Added new id to ou: %s, type: %s, source: %s value: %s",
                ou_id,
                identity_type,
                source,
                identity_in_orgreg_value,
            )

    def _get_or_create_and_set_values(
        self, ou: OrgUnit, values: Mapping[str, Union[str, int, bool]]
    ):
        """Upsert ou with latest values and store in processed dict."""
        try:
            identifier = OuIdentifier.objects.get(
                name=settings.ORGREG_NAME, source=settings.ORGREG_SOURCE, value=ou.ou_id
            )
        except OuIdentifier.DoesNotExist:
            self.processed[ou.ou_id] = OrganizationalUnit.objects.create()
            identifier = OuIdentifier.objects.create(
                name=settings.ORGREG_NAME,
                source=settings.ORGREG_SOURCE,
                value=ou.ou_id,
                orgunit=self.processed[ou.ou_id],
            )
            created = True
        else:
            self.processed[ou.ou_id] = identifier.orgunit
            created = False
        for k, v in values.items():
            setattr(self.processed[ou.ou_id], k, v)
        self._upsert_extra_identities(ou)
        self.processed[ou.ou_id].save()
        logger.info(
            "%s %s with %s",
            "Created" if created else "Updated",
            self.processed[ou.ou_id],
            values,
        )

    def _upsert_ou(self, ou: OrgUnit, ous: Mapping[int, OrgUnit]):
        """
        Update or create a OU with current values.

        There are three cases:
        1. OU has no parent -> Create it.
        2. OU has parent and parent exists -> Create it.
        3. OU has parent but parent has not been created -> Create parent, then child.

        Case 3 is solved by calling this method recursively so that the parent and its
        parent are created before the child.
        """

        # skip if already processed (happens if a previous ou needed this one as a
        # parent)
        if ou.ou_id in self.processed:
            return

        values = {"deleted": False}
        # add names if present
        if ou.name:
            if ou.name.nob:
                values["name_nb"] = ou.name.nob
            if ou.name.eng:
                values["name_en"] = ou.name.eng

        # Set inactive if ou is no longer valid
        is_inactive = ou.valid_to and ou.valid_to < datetime.date.today()
        values["active"] = not is_inactive

        # Create OUs without parent (this should only happen for the root node)
        if not ou.parent:
            self._get_or_create_and_set_values(ou, values)
            return

        # If OU has parent and it is not already created, we create it and point the
        # child to it.
        if ou.parent not in self.processed:
            self._upsert_ou(ous[ou.parent], ous)
        values["parent"] = self.processed[ou.parent]

        # Finally create the OU we're currently looking at
        self._get_or_create_and_set_values(ou, values)

    def handle(self, *args, **options):
        """
        Handle import of OUs from OrgReg.

        - Updates already present OUs with new information
        - Creates not present OUs
        - Set not present OUs as inactive
        """
        # Empty processed in case of class reuse by tests
        self.processed = {}
        client = orgreg_client.get_client(**settings.ORGREG_CLIENT)

        # Fetch already present OUs and those in OrgReg
        current_ous = {
            int(i.value): i.orgunit
            for i in OuIdentifier.objects.filter(
                source=settings.ORGREG_SOURCE, name=settings.ORGREG_NAME
            )
        }

        logger.info("Fetch OUs from Orgreg...")
        orgreg_ous = {i.ou_id: i for i in client.get_ou()}

        # Set deleted if an OU from Greg no longer exists in OrgReg
        logger.info("Set deleted tag on removed OUs...")
        for ou_id, ou in current_ous.items():
            if ou_id not in orgreg_ous:
                logger.info("%s marked as deleted", ou)
                ou.deleted = True
                ou.save()
                continue

        # Create new OUs recursively to ensure parents exist before children
        logger.info("Update OU values...")
        for ou_id, ou in orgreg_ous.items():
            self._upsert_ou(ou, orgreg_ous)
