from rest_framework import mixins
from rest_framework.authentication import SessionAuthentication, BasicAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.viewsets import GenericViewSet
from greg.api.serializers.person import PersonSearchSerializer, SpecialPersonSerializer

from greg.models import Identity, Person
from greg.permissions import IsSponsor
from gregui import validation
from gregui.api.serializers.guest import create_identity_or_update
from gregui.models import GregUserProfile


class PersonViewSet(mixins.RetrieveModelMixin, mixins.UpdateModelMixin, GenericViewSet):

    """
    Fetch person info for any guest as long as you are a sponsor

    This is required for the functionality where a sponsor wants to add a guest role to
    an already existing person that is not already their guest.

    Returns enough information to fill a profile page in the frontend
    """

    authentication_classes = [SessionAuthentication, BasicAuthentication]
    permission_classes = [IsAuthenticated, IsSponsor]
    queryset = Person.objects.all()
    http_methods = ["get", "patch"]
    serializer_class = SpecialPersonSerializer

    def perform_update(self, serializer):
        """Update email when doing patch"""
        email = self.request.data.get("email")
        person = self.get_object()
        validation.validate_email(email)
        create_identity_or_update(Identity.IdentityType.PRIVATE_EMAIL, email, person)
        return super().perform_update(serializer)


class PersonSearchViewSet(mixins.ListModelMixin, GenericViewSet):
    """Search for persons using email or phone number"""

    authentication_classes = [SessionAuthentication, BasicAuthentication]
    permission_classes = [IsAuthenticated, IsSponsor]
    serializer_class = PersonSearchSerializer

    def get_queryset(self):
        search = self.kwargs["searchstring"]
        return Identity.objects.filter(
            value__icontains=search,  # icontains to include wrong case emails
            type__in=[
                Identity.IdentityType.PRIVATE_EMAIL,
                Identity.IdentityType.PRIVATE_MOBILE_NUMBER,
            ],
        )[:10]


class GuestInfoViewSet(mixins.ListModelMixin, GenericViewSet):
    """
    Fetch all guests at the sponsor's units.

    If the Sponsor's connection to the Unit is marked with hierarchical access, guests
    at child units of the Unit are included.

    Lists all persons connected to the roles the logged in sponsor is connected to.
    """

    authentication_classes = [SessionAuthentication, BasicAuthentication]
    permission_classes = [IsAuthenticated, IsSponsor]
    serializer_class = SpecialPersonSerializer

    def get_queryset(self):
        """
        Fetch Persons connected to the sponsor some way.

        Any person with a role connected to the same unit as the sponsor, or a unit
        that the sponsor is connected to through hierarchical access.
        """

        user = GregUserProfile.objects.get(user=self.request.user)
        units = user.sponsor.get_allowed_units()
        return (
            Person.objects.filter(roles__orgunit__in=list(units))
            .distinct()
            .order_by("id")
        )
